#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2005 by S.Çağlar Onur <caglar@uludag.org.tr>
#
# InitNG RunLevel Editor
# For more information about InitNG go to:
# http://initng.thinktux.net/index.php/Main_Page
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from qt import *
from kdeui import *
from kdecore import *
from dcopexport import DCOPExObj

''' import UI '''
from initng import *

''' Standart Python Modules '''
import os
import sys
import glob
import re
from itertools import ifilter

''' Import Gettext Support '''
import gettext
__trans = gettext.translation('InitNG Runlevel Editor', fallback=True)
_  =  __trans.ugettext


# Helper functions/singletons...

def cat(filename):
    ''' cat file same as cat command '''
    return file(filename).xreadlines()

class grep:
    ''' keep only lines that match the regexp '''
    def __init__(self, pattern, flags = 0):
        self.fun = re.compile(pattern, flags).match
    def __ror__(self, input):
        return ifilter(self.fun, input)

class uniq:
    ''' remove duplicates same as uniq command '''
    def __ror__(self, input):
        for i in input:
            try:
                if i == prev:
                    continue
            except NameError:
                pass
            prev = i
            yield i
uniq = uniq()

class terminator:
    def __init__(self,method):
        self.process = method
    def __ror__(self,input):
        return self.process(input)

aslist = terminator(list)

class InitNG_Runlevel_Editor(InitNG_Runlevel_EditorImpl):

    # Globals        
    runlevelName = ["default", "system", "single"]
    scriptsDirectory = ["daemon", "daemon/bluetooth", "system", "net"]

    def __init__(self, *args):
        ''' constuctor '''
        apply(InitNG_Runlevel_EditorImpl.__init__, (self,) + args)

        ''' DCOP Interface '''
        self.DCOPInstance = DCOPIface()

        # KActionSelector have 2 ListBox 
        self.availableListBox = self.actionSelector.availableListBox()
        self.selectedListBox = self.actionSelector.selectedListBox()
        
        # SIGNAL/SLOT...

        # ComboBox changed
        self.connect(self.comboBox, SIGNAL("activated(int)"), self.fillWithRunlevel)
        self.connect(self.comboBox, SIGNAL("activated(int)"), self.fillWithScripts)

        # Apply button pressed        
        self.connect(self.applyButton, SIGNAL("clicked()"), self.applyChanges)

        # Quit button pressed
        self.connect(self.quitButton, SIGNAL("clicked()"), self.close)

        # Check uid
        if not os.geteuid() == 0:
            self.applyButton.setEnabled(0)
            QMessageBox.information(self, "Unsufficient Permission", "You must run as root", QMessageBox.Ok)

        # Init...
        self.fillWithScripts()
        self.fillWithRunlevel(0)

    def fillWithScripts(self):
        ''' fill availableListBox with all initscripts not in selected runlevel '''
        self.availableListBox.clear()

        for currentDirectory in self.scriptsDirectory:
            for currentFile in glob.glob("/etc/initng/%s/*.i" % currentDirectory):
                file = "%s/%s" % (currentDirectory, os.path.basename(currentFile))
                # remove .i extension
                file = os.path.splitext(file)[0]
                if not self.selectedListBox.findItem(file):
                    self.availableListBox.insertItem(file)

        self.availableListBox.sort()

    def fillWithRunlevel(self, runlevelID):
        ''' fill selectedListBox with initscripts in current selected runlevel '''
        self.selectedListBox.clear()
        
        for line in cat("/etc/initng/%s.runlevel" % self.runlevelName[runlevelID]) | uniq | aslist:
            line = line.strip()
            self.selectedListBox.insertItem(line)

        self.selectedListBox.sort()

    def applyChanges(self):
        ''' apply changes and save file '''
        try:
            runlevelFile = open("/etc/initng/%s.runlevel" % self.runlevelName[self.comboBox.currentItem()], "w")

            for index in range(self.selectedListBox.count()):
                runlevelFile.write("%s\n" % self.selectedListBox.text(index))

            runlevelFile.close()

            QMessageBox.information(self, "Done", "%s runlevel updated..." % self.runlevelName[self.comboBox.currentItem()], QMessageBox.Ok)
        except IOError, error:
            QMessageBox.critical(self, "Error Occured", "%s" % error, QMessageBox.Ok)

class DCOPIface(DCOPExObj):
    def __init__(self):
        ''' Initialize DCOP '''
        DCOPExObj.__init__(self, "InitNGIface")

        self.addMethod("QStringList returnDefaultRunlevel()", self.returnDefaultRunlevel)
        self.addMethod("QStringList returnSystemRunlevel()", self.returnSystemRunlevel)
        self.addMethod("QStringList returnSingleRunlevel()", self.returnSingleRunlevel)
        
    def returnRunlevelList(self, runlevelName = "default"):
        ''' return given runlevel's list '''
        list = QStringList()
        for line in cat("/etc/initng/%s.runlevel" % runlevelName) | uniq | aslist:
            list.append(line.strip())
        return list
    
    def returnDefaultRunlevel(self):
        ''' return default runlevel's list '''
        return self.returnRunlevelList()

    def returnSystemRunlevel(self):
        ''' return system runlevel's list '''
        return self.returnRunlevelList("system")
    
    def returnSingleRunlevel(self):
        ''' return single runlevel's list '''
        return self.returnRunlevelList("single")

    def addToRunlevel(self, runLevelName = "default"):
        pass

    def removeFromRunlevel(self, runLevelName = "default"):
        pass

if __name__ == '__main__':
    appName     = 'InitNG'
    programName = 'InitNG Runlevel Editor'
    description = 'InitNG Runlevel Editor'
    license     = KAboutData.License_GPL
    version     = '0.1'
    copyright   = '(C) 2005 S.Çağlar Onur <caglar@uludag.org.tr>'

    aboutData = KAboutData(appName, programName, version, description, license, copyright)
    aboutData.addAuthor('S.Çağlar Onur', 'Developer', 'caglar@uludag.org.tr')

    KCmdLineArgs.init(sys.argv, aboutData)
    
    ''' i18n support '''
    gettext.install(appName)

    ''' Use KUniqueApplication and initialize '''
    app = KUniqueApplication(True, True, True)

    QObject.connect(app, SIGNAL("lastWindowClosed()"), app, SLOT("quit()"))    
    
    ''' Create main window '''
    mainWindow = InitNG_Runlevel_Editor()
    app.setMainWidget(mainWindow)
    mainWindow.show()

    ''' Enter main loop '''
    app.exec_loop()
