/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"
#include "../../src/initng_handler.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "../../src/initng_global.h"
#include "../../src/initng_string_tools.h"
#include "../../src/initng_active_db.h"
#include "../../src/main.h"
#include "../../src/initng_plugin_hook.h"
#include "initng_colorprint_out.h"

#define PE C_RED " %3i%% " C_OFF C_BLUE "%s" C_OFF MOVE_TO_R
#define P C_RED "      " C_OFF C_BLUE "%s" C_OFF MOVE_TO_R

active_h *lastservice;
static void clear_lastserv(void)
{
    if (lastservice)
        printf("\n");
    lastservice = NULL;
}

static void out_service_done(active_h * s)
{
    int t;
    int percent;

    assert(s);
    assert(s->name);
    t = time(0) - s->uptime;
    percent = active_db_percent_started();

    clear_lastserv();
    if (t > 0)
        printf(PE "       [" C_GREEN "done" C_OFF
               "]   ( started in %i second%s. )\n", percent, s->name, t,
               (t == 1) ? "" : "s");
    else
        printf(PE "       [" C_GREEN "done" C_OFF "]\n", percent, s->name);
}

static void opt_service_stop_p(const char *status, char *name)
{
    assert(status);
    assert(name);
    clear_lastserv();
    printf(PE "       [" C_GREEN "%s" C_OFF "]\n",
           active_db_percent_stopped(), name, status);
}

/*
   static void out_service(const char *status, char *name)
   {
   assert(status);
   assert(name);
   clear_lastserv();
   printf(P "    [" C_GREEN "%s" C_OFF "]\n", status, name);
   } */

int print_output(active_h * service)
{
    assert(service);
    assert(service->name);

    /* if, do some output and history adds */
    switch (service->a_status)
    {
        case STOPPED:
            opt_service_stop_p("stopped", service->name);
            break;
        case STARTING:
            /* if we print this on boot, we clutter up the screen to mutch */
            if (g.sys_state == STATE_STARTING)
                break;
            clear_lastserv();
            printf(PE "   [" C_GREEN "starting" C_OFF "]\n",
                   active_db_percent_started(), service->name);
            break;
        case DONE:
            out_service_done(service);
            break;
        case RUNNING:
            {
                process_h *process = NULL;

                while_processes(process, service)
                {
                    if (process->pt != T_DAEMON)
                        continue;

                    clear_lastserv();
                    printf(PE "    [" C_GREEN "started" C_OFF
                           "]       (pid : %i).\n",
                           active_db_percent_started(), service->name,
                           process->pid);
                }
            }
            break;
        case STOP_MARKED:
            opt_service_stop_p("stopping", service->name);
            break;
        default:
            break;
    }
    return (TRUE);
}


static void print_system_state(h_sys_state state)
{
    switch (state)
    {
        case STATE_STARTING:
            clear_lastserv();
            printf("    system is starting up!\n\n");
            break;
        case STATE_STOPPING:
            clear_lastserv();
            printf("    system is going down!\n\n");
            break;
        case STATE_ASE:
            clear_lastserv();
            printf("    Last process exited!\n\n");
            break;
        case STATE_EXIT:
            clear_lastserv();
            printf("    Initng exiting!\n\n");
            break;
        case STATE_RESTART:
            clear_lastserv();
            printf("    Initng restarting!\n\n");
            break;
        case STATE_UP:
            {
                active_h *runl = active_db_find_by_name(g.runlevel);

                if (!runl)
                    return;
                int t = time(0) - runl->uptime;

                clear_lastserv();

                if (t > 3)
                    printf
                        ("\n\n       runlevel \"%s\" up in, %i second%s.\n\n",
                         g.runlevel, t, (t == 1) ? "" : "s");
                break;
            }
        default:
            break;
    }
    D_("print_system_state(): new system state: %i\n", state);
    return;
}

void print_program_output(active_h * service, int datalen, process_h * x)
{
    /*
       Todo here:
       we should have an internal list of services and "our" current position in them.

       That way when this function is called we can print every full line from plugin_pos.
       This way fsck will look nice, along with an "internal" database of write positions we can cache data so we print every 5 seconds or on int forceflush.
     */
    /*int skipline = FALSE; */
    int i;

    assert(service);
    assert(service->name);
    assert(x);

    S_;
    D_(" from service \"%s\"\n", service->name);

    if (datalen == 0)
    {
        D_("service %s datalen == 0\n", service->name);
        return;
    }
    /*
       printf("buffer_pos: %i\n", x->buffer_pos);
       printf("datalen: %i\n", datalen);
       printf("Buffer: \n################\n%s\n##########\n\n",x->buffer);
     */
    i = x->buffer_pos;

    /* a first while loop that sorts out crap */
    while (x->buffer[i] != '\0')
    {
        /*  remove lines with " [2]  Done " that bash generates. */
        if (x->buffer[i] == '[' && x->buffer[i + 2] == ']')
        {
            /* jump to next line */
            while (x->buffer[i] && x->buffer[i] != '\n')
                i++;
        }

        /* if there are stupid tockens, go to next char, and run while again. */
        if (x->buffer[i] == ' ' || x->buffer[i] == '\n'
            || x->buffer[i] == '\t')
        {
            i++;
            continue;
        }

        /* else break */
        break;
    }

    /* Make sure that there is anything left to write */
    if (strlen(&x->buffer[i]) < 2)
    {
        return;
    }

    if (lastservice != service)
    {
        clear_lastserv();
        printf("\n" C_CYAN " %s:" C_OFF, service->name);
        /* print our special indented newline */
        putchar('\n');
        putchar(' ');
        putchar(' ');
        lastservice = service;

    }
    else
    {
        D_("Lastservice == service, won't print header.\n");
    }



    /* while buffer lasts */
    while (x->buffer[i] != '\0')
    {
        /*  remove lines with " [2]  Done " that bash generates. */
        if (x->buffer[i] == '[' && x->buffer[i + 2] == ']')
        {
            while (x->buffer[i] && x->buffer[i] != '\n')
                i++;
        }

        /* if this are a newline */
        if (x->buffer[i] == '\n')
        {
            /* print our special indented newline instead */
            putchar('\n');
            putchar(' ');
            putchar(' ');
            i++;

            /* skip spaces, on newline. */
            while (x->buffer[i]
                   && (x->buffer[i] == ' ' || x->buffer[i] == '\t'))
                i++;
            continue;
        }

        /* ok, now put the char, and go to next. */
        putchar(x->buffer[i]);
        i++;
    }
}


static int cp_print_error(e_mt mt, const char *file, const char *func,
                          int line, const char *format, va_list arg)
{
    assert(file);
    assert(func);

    struct tm *ts;
    time_t t;



    switch (mt)
    {
        case MSG_FAIL:
            t = time(0);
            ts = localtime(&t);

            fprintf(stderr, "\n\n ** \"%s\", %s():\n", file, func);
            fprintf(stderr, " %.2i:%.2i:%.2i -- line:%i FAIL:\t", ts->tm_hour,
                    ts->tm_min, ts->tm_sec, line);
            vfprintf(stderr, format, arg);
            break;
        case MSG_WARN:
            t = time(0);
            ts = localtime(&t);

            fprintf(stderr, "\n\n ** \"%s\", %s():\n", file, func);
            fprintf(stderr, " %.2i:%.2i:%.2i -- line:%i WARN:\t", ts->tm_hour,
                    ts->tm_min, ts->tm_sec, line);
            vfprintf(stderr, format, arg);

            break;
        default:
            vfprintf(stderr, format, arg);
            break;
    }
    return (TRUE);
}


int module_init(const char *version)
{

    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    printf("\n" C_BLUE "     Next Generation Init version ( %s )" C_OFF,
           INITNG_VERSION);
    printf("\n" C_GREEN "     http://initng.thinktux.net\n" C_OFF);
    printf("     Author: Jimmy Wennlund <jimmy.wennlund@gmail.com>\n");
    printf
        ("     If you find initng useful, please consider a small donation.\n\n");

    D_("module_init();\n");
    lastservice = NULL;
    initng_add_hook(ERR_MSG, 10, &cp_print_error);
    initng_add_hook(ASTATUS_CHANGE, 80, &print_output);
    initng_add_hook(SWATCHERS, 80, &print_system_state);
    initng_add_hook(PIPEWATCHERS, 50, &print_program_output);
    return (TRUE);
}

void module_unload(void)
{
    D_("color_out: module_unload();\n");
    initng_del_hook(ASTATUS_CHANGE, &print_output);
    initng_del_hook(SWATCHERS, &print_system_state);
    initng_del_hook(PIPEWATCHERS, &print_program_output);
    initng_del_hook(ERR_MSG, &cp_print_error);
    printf("  Goodbye\n");
}
