
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_open_read_close.h"
#include "../../src/initng_string_tools.h"

#include "initng_envparser.h"
s_entry ENV_FILE = { "env_file", STRING, 8, STD_OPT };


static char **env_parser(active_h * s, char **env, int alloced)
{
    int i = 0;
    int j = 0;
    char *file = NULL;
    char *file_content = NULL;
    char *point = NULL;
    
    assert(s);

    /* go to last row on this array */
    while (env[i])
        i++;

    /* get filename */
    file = active_db_get_string(&ENV_FILE, s);
    if (!file)
        return (env);

    /* open that file */
    if (!open_read_close(file, &file_content))
        return (env);


    /* Start Parsing */
    point = file_content;
    while (point[0])
    {
        /* skip initzial spaces */
        JUMP_SPACES(point);
        if (!point[0])
            break;

        /* skip rows, starting with '#' */
        if (point[0] == '#')
        {
            JUMP_TO_NEXT_ROW(point);
            continue;
        }

        /* make sure the line contains a '=', if not, its not an ENV line */
        j = 0;
        while (point[j] && point[j] != '=' && point[j] != '\n')
            j++;
        if (point[j] != '=')
        {
            JUMP_TO_NEXT_ROW(point);
            continue;
        }

        /* duplicate this line */
        env[i] = st_dup_line(&point);
        i++;
        /* make sure last line, is always NULL */
        env[i] = NULL;

        /* Do a check that we don't exeeds the max no of env we can add */
        if (i >= (alloced - 1))
        {
            free(file_content);
            return (env);
        }

    }
    free(file_content);
    return (env);
}

int module_init(const char *version)
{

    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }
    D_("module_init();\n");

    initng_sdt_add(&ENV_FILE);
    return (initng_add_hook(NEW_ENVIRON, 50, &env_parser));
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_sdt_del(&ENV_FILE);
    initng_del_hook(NEW_ENVIRON, &env_parser);
}
