
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"
#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <dirent.h>                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                /*#include <dirent.h> */     /* opendir() closedir() */
#include <fnmatch.h>

#include "../../src/initng_string_tools.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_global.h"
#include "../../src/initng_open_read_close.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"

#include "initng_i_parser.h"

static void print_line(char *from);
static service_h *parse_file(char *filename, char *watch_for);
static int parse_service_line(char **to_parse, char *watch_for,
                                     service_h * father, e_type type,
                                     char *filename, service_h **match, service_h **exact_match);
static int parse_opt(char **where, e_type type, service_h * srv);

static int set_parser(s_entry * type, char **value, service_h * from_service);
static int unset_parser(s_entry * type, char **value,
                        service_h * from_service);
static int string_parser(s_entry * type, char **value,
                         service_h * from_service);
static int strings_parser(s_entry * type, char **value,
                          service_h * from_service);
static int script_parser(s_entry * type, char **value,
                         service_h * from_service);
static int int_parser(s_entry * type, char **value, service_h * from_service);



/* Load a service from a service_to_find or process_path */
service_h *initng_i_parser(char *service_to_find)
{
    char filetoparse[200];
    char *tmp = NULL;
    service_h *got_serv = NULL;

    assert(service_to_find);
    D_("initng_i_parser(%s);\n", service_to_find);
    /* example  INITNG_ROOT=/etc/initng service_to_find="deamon/sshd INITNG_EXT=".i" */
    /* TRY NO 1, try load   "/etc/initng/daemon/sshd.i" */
    /* TRY NO 2, try load   "/etc/initng/daemon/sshd/sshd.i" */
    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.i" */
    /* TRY NO 4, try load   "/etc/initng/daemon/default.i" */
    /* TRY NO 5, try load   "/etc/initng/daemon/daemon.i */
    /* TRY NO 6, try load   "/etc/initng/sshd.i" */
    /* TRY NO 7, try load   "/etc/initng/daemon.i */

    /* TRY NO 1, try load   "/etc/initng/daemon/sshd/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, service_to_find);
    strcat(filetoparse, "/");
    strcat(filetoparse, st_strip_path(service_to_find));
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
        return (got_serv);

    /* TRY NO 2, try load   "/etc/initng/daemon/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, service_to_find);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
        return (got_serv);

    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, service_to_find);
    strcat(filetoparse, "/default" INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
        return (got_serv);

    /* TRY NO 4, try "/etc/initng/daemon/default.i" */
    tmp = st_get_path(service_to_find);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, "/default" INITNG_EXT);
    free(tmp);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
        return (got_serv);

    /* TRY NO 5, try "/etc/initng/daemon/daemon.i" */
    tmp = st_get_path(service_to_find);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, INITNG_EXT);
    free(tmp);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
        return (got_serv);

    /* TRY NO 6, try "/etc/initng/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, st_strip_path(service_to_find));
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
        return (got_serv);

    /* TRY NO 7, try "/etc/initng/daemon.i" */
    tmp = st_get_path(service_to_find);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, INITNG_EXT);
    free(tmp);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
        return (got_serv);


    D_("Did not be able to parse: %s\n", service_to_find);
    return (NULL);
}


static void print_line(char *from)
{
    /* Complain a bit */
    char *line = i_strndup(from, strcspn(from, "\n\0"));
    D_("%s\n", line);
    free(line);
}

/* parse a file for services */
static service_h *parse_file(char *filename, char *watch_for)
{
    char *file, *orig;
    e_type t=TYPE_UNKNOWN;
    service_h *match = NULL;
    service_h *exact_match = NULL;

    assert(filename);
    assert(watch_for);

    D_("parse_file(%s);\n", filename);

    /* read config file */
    if (!open_read_close(filename, &orig))
    {
        D_("Can't open config file %s.\n", filename);
        return (NULL);
    }
    file = orig;


    /* main parse line by line loop */
    while (file[0] != '\0')
    {
        /* skip leading spaces */
        JUMP_SPACES(file);
        if (file[0] == '\0')
            break;

        /* skip lines starting with '#' */
        if (file[0] == '#' || file[0] == '\n' || file[0] == ';')
        {
            JUMP_TO_NEXT_ROW(file);
            continue;
        }

        /* if this row has keyword service */
        if (st_cmp(&file, "service"))
            t=TYPE_SERVICE;
        else if (st_cmp(&file, "class"))
	    t=TYPE_CLASS;
        else if (st_cmp(&file, "daemon"))
	    t=TYPE_DAEMON;
        else if (st_cmp(&file, "virtual"))
	    t=TYPE_VIRTUAL;
	

	if(t!=TYPE_UNKNOWN)
	{
	    if(parse_service_line(&file, watch_for, NULL, t,
                                    filename, &match, &exact_match)==TRUE)
	    {
		/* if we found RIGHT what we wanted */
		if(exact_match)
		{
		    free(orig);
		    return(exact_match);
		}
		    
	        JUMP_TO_NEXT_ROW(file);
		continue;
	    }
	}

	/* else BAIL OUT */
		
    	F_("PARSE ERROR parse_file(%s): ", watch_for);
	print_line(file);
    	F_("sleeping 2 seconds, then ABORT......\n");
    	sleep(2);
	free(orig);
	return(NULL);
    }                                       /* end while */

    free(orig);

    /* Return what we got, and hope for the best */
    return (match);
}




static int parse_service_line(char **to_parse, char *watch_for,
                                     service_h * father, e_type type,
                                     char *filename, service_h **match, service_h **exact_match)
{
    char *name = NULL;
    char *father_name = NULL;
    service_h *new_service = NULL;
    service_h *duplicate = NULL;

    assert(to_parse);
    assert(*to_parse);
    assert(watch_for);

    D_("parse_service_line(%s): ", watch_for);
    print_line(*to_parse);

    /* service test : class { */
    /*        |               */
    /* jump forward on spaces */
    JUMP_SPACES(*to_parse);

    /* service test : class { */
    /*         |              */

    /* fetch name */
    if (!(name = st_dup_next_word(to_parse)))
    {
        F_("Did not got name!\n");
        return (FALSE);
    }
    /* service test : class { */
    /*             |          */

    D_("parse_service_line(s,%s): service name: \"%s\"\n", watch_for, name);

    /* jump forward on spaces */
    JUMP_SPACES(*to_parse);

    /* service test : class { */
    /*              |         */
    if ((*to_parse)[0] == ':')
    {
        (*to_parse)++;
        /* service test : class { */
        /*               |        */

        /* jump forward on spaces */
        JUMP_SPACES(*to_parse);
        /* service test : class { */
        /*                |       */

        father_name = st_dup_next_word(to_parse);
        if (father_name)
        {
            D_("Father is: %s, set from \"server : father\"\n", father_name);
        }

        /* service test : class { */
        /*                     |  */
        /* jump forward on spaces */
        JUMP_SPACES(*to_parse);

    }

    /* service test : class { */
    /*                      | */
    if ((*to_parse)[0] != '{')
    {
        /* something is wrong */
        F_("error parsing, expected something else.\n");
        return (FALSE);
    }

    /* jump to first char after start tag and begin from there. */
    (*to_parse)++;
    /* service test : class { */
    /*                       | */

    /* create a service entry - if there exits an duplicate, merge with this one */
    if ((duplicate = service_db_find_by_exact_name(name)))
        new_service = service_db_copy(name, duplicate);
    else
        new_service = service_db_new(name, type);

    /* check so that it was allocated! */
    if (!new_service)
    {
        F_("Unable to allocate space for new service!\n");
        /* need to free any allocated strings */
        free(name);
        if (father_name)
            free(father_name);
        return (FALSE);
    }

    /* set the father to the service */
    if (father_name)
    {
        new_service->father_name = father_name;
        /* father will be service_db_find_by_name(father_name) when needed */
    }
    else if (father && father->name)
    {
        new_service->father_name = father->name;
        new_service->father = father;
    }

    /* why another check? if new_service not exist it will fail upper */
    /*    if (!new_service)
       {
       F_("parse_service_line(s,s): unable to allocate new service!\n");
       free(name);
       if (father_name)
       free(father_name);
       return (NULL);
       }
     */
    if (duplicate)
    {
        /* Service has same name as old service, remove the old one */
        D_("Removing old service for %s\n", new_service->name);
        /* remove it from the service_db */
        list_del(&duplicate->list);
        /* free all the data, and service */
        service_db_free(duplicate);
    }

    /* set type if not set. */
    if (father)
        new_service->type = type;


    /* carry on until segment stop or eof, this will handle all lines in current section */
    while ((*to_parse)[0] != '\0' && (*to_parse)[0] != '}')
    {
        while (TRUE)
        {
            /* skip spaces and empty lines with them */
            JUMP_SPACES(*to_parse);
            if (*(to_parse)[0] == '\0' || *(to_parse)[0] == '}')
                break;
            if (*(to_parse)[0] == '\n')
                (*to_parse)++;
            else
                break;
        }

        /* skip lines with '#' */
        if ((*to_parse)[0] == '#')
        {
            JUMP_TO_NEXT_ROW(*to_parse);
            continue;
        }

        /****  Check for subclasses ****/

        /* if this row has keyword service */
        if (st_cmp(to_parse, "service") == TRUE)
        {
    	    if(parse_service_line(to_parse, watch_for, NULL, TYPE_SERVICE,
                                    filename, match, exact_match)==FALSE)
		return(FALSE);
            JUMP_TO_NEXT_ROW(*to_parse);
            continue;
        }

        /* if this row has keyword daemon */
        /*
           if (st_cmp(to_parse, "daemon") == TRUE)
           {
           if ((tmp =
           parse_service_line(to_parse, watch_for, NULL, TYPE_DAEMON,
           filename)))
           service_to_return = tmp;
           JUMP_TO_NEXT_ROW(*to_parse);
           continue;
           } */

        /* if this row has keyword class */
        if (st_cmp(to_parse, "class") == TRUE)
        {
            if (parse_service_line(to_parse, watch_for, NULL, TYPE_CLASS,
                                    filename, match, exact_match)==FALSE)
		return(FALSE);
            JUMP_TO_NEXT_ROW(*to_parse);
            continue;
        }

        /* parse line - search for keywords in g.option_table */
        if(!parse_opt(to_parse, type, new_service)) {
	    if(new_service) service_db_free(new_service);
	    return(FALSE);
	}
	
        /* ok, this line is parsed, go to next */
        JUMP_TO_NEXT_ROW(*to_parse);
    }

    service_db_set_string(&FROM_FILE, new_service, i_strdup(filename));
    /* add to service db */
    if (!service_db_add(new_service))
    {
        F_("can't add service to service db!\n");
        if (new_service->name)
            free(new_service->name);

        /* remove all data entrys */
        service_db_remove_all(new_service);

        if (new_service)
            service_db_free(new_service);
        return (FALSE);
    }
    
    if(strcmp(new_service->name, watch_for)==0)
    {
	D_("Found EXACT match: %s\n", new_service->name);
	(*exact_match)=new_service;
	return(TRUE);
    }
    
    /* check if this is the service we are looking for. */
    if (service_match(watch_for, new_service->name)==TRUE)
    {
	D_("Found :%s\n", new_service->name);
	(*match)=new_service;
	return(TRUE);
    }
    
    D_("No match\n");
    /* return null if not */
    return (TRUE);
}



static int parse_opt(char **where, e_type type, service_h * srv)
{
    s_entry *current = NULL;
    char *opt_name = NULL;
    int opt_len=0;
    int w_l = 0;
    int result = FALSE;
    int multiline = FALSE;
    D_("parse_opt: ");
    print_line(*where);

    while (TRUE)
    {
        /* jump forward all spaces and skip empty or commented lines after that */
        JUMP_SPACES(*where);
        if ((*where)[0] == '\n')
            (*where)++;
        else
            break;
    }

    /* skip empty services */
    if ((*where)[0] == '}')
        return (TRUE);

    /* count number of chars this option name have 
       (option name can contain blank characters - i.e. spaces and tabs */
    while ((*where)[w_l] && (*where)[w_l] != '\n'
           && (*where)[w_l] != ';' && (*where)[w_l] != '=')
        w_l++;

    /* strip ' ' and '\t' from the end */
    while ((*where)[w_l - 1] == ' ' || (*where)[w_l - 1] == '\t')
        w_l--;

    if (!w_l)
    {
        D_("option name to short!\n");
        return (FALSE);
    }

    /* detect if it is a multi line option */
    {
	int i;
	for(i=0; i<w_l; i++)
	{
	    switch((*where)[i])
	    {
		/* this idicates that it ends */
		case '\n':
		case ';':
		case '\0':
		    multiline=FALSE;
		    i=w_l+1; /* stops the for loop */
		    break;
		/* this indicates that option will be in more rows */
		case '{':
		    multiline=TRUE;
		    i=w_l+1; /* stops the for loop */
		    break;
	    }
	}
    }
	
    /* get option name and strip all blanks from it */
    opt_name = initng_calloc(w_l+2, sizeof(char));
    {
	int i;
	int j=0;
	for(i=0; i<=w_l; i++)
	{
	    switch((*where)[i])
	    {
		case ' ':
		case '\t':
		    continue;
		case '\n':
		case ';':
		    i=w_l+1;
		    continue;
	    }
	    opt_name[j++]=(*where)[i];
	}
	opt_name[j]='\0';
    }
    /* we save this, so we dont need to count this every while below */
    opt_len=strlen(opt_name);
    
    /* walk the option db */
    while_option_db(current)
    {
        /* temporary pointer storage */
        s_entry *tmp = current;

        /* make sure opt_name is set */
        if (!current->opt_name)
            continue;

        /* check that option is right type */
        switch (type)
        {
            case TYPE_SERVICE:
                if (current->ot != STD_OPT && current->ot != SERVICE_OPT)
                    continue;
                break;

            case TYPE_DAEMON:
                if (current->ot != STD_OPT && current->ot != DAEMON_OPT)
                    continue;
                break;

                /* TODO, how shud we check this data? */
            case TYPE_CLASS:
                break;

            case TYPE_VIRTUAL:
                if (current->ot != STD_OPT)
                    continue;
                break;

                /* dont ever parse other types */
            default:
                continue;
        }

        /* check string lenght of word, this is faster then strcasecmp */
        if (opt_len != current->str_len)
            continue;

        /* finaly, check if opt_name is what we are looking for */
        if (strncmp(current->opt_name, opt_name, opt_len) != 0)
            continue;

        D_("parse_opt(%s): option \"%s\" value at option.\n",
           srv->name, current->opt_name);

        /* jump one char */
        *where = *where + w_l;
        /* jump forward all spaces */
        JUMP_SPACES(*where);

        /* check that we are not in end of line */
        if (!(*where)[0])
            continue;

        /* if this is a alias, browse forward to the correct opt_type */
        while (tmp->opt_type == ALIAS && tmp->alias)
            tmp = tmp->alias;

	if(multiline)
	{
	    if(tmp->opt_type==MULTILINE_STRING) {
                result = script_parser(current, where, srv);
	    } else {
		D_("Option is contains a '{' but is not a MULTILINE_STRING!\n");
		continue;
	    }
	} else {
        /* now switch the opt_type */
        switch (tmp->opt_type)
        {
            case STRING:
                result = string_parser(current, where, srv);
                break;
            case STRINGS:
                result = strings_parser(current, where, srv);
                break;
            case MULTILINE_STRING:
                result = script_parser(current, where, srv);
                break;
            case SET:
                result = set_parser(current, where, srv);
                break;
            case UNSET:
                result = unset_parser(current, where, srv);
                break;
            case INT:
                result = int_parser(current, where, srv);
                break;
            default:
                F_("Unknowun, or undefined type to parse!\n");
                continue;
        }
	}
    }

    free(opt_name);

    /* return nicely */
    if (result)
        return (result);


    /* Complain a bit */
    {
        char *line = i_strndup(*where, strcspn(*where, "\n\0"));

        F_("parse_opt(%s): line: \"%s\"!\n", srv->name, line);
        free(line);
        F_("sleeping 2 seconds......\n");
        sleep(2);
    }
    return (FALSE);
}



static int set_parser(s_entry * type, char **value
                      __attribute__ ((unused)), service_h * from_service)
{
    assert(from_service);
    assert(*value);
    assert(value);

    D_("set_parser(%s,s,%s);\n", type->opt_name, from_service->name);
    service_db_set(type, from_service);
    return (TRUE);
}

static int unset_parser(s_entry * type, char **value __attribute__ ((unused)),
                        service_h * from_service)
{
    assert(from_service);
    assert(*value);
    assert(value);

    D_("unset_parser(%s,s,%s);\n", type->opt_name, from_service->name);
    service_db_remove(type, from_service);
    return (TRUE);
}

static int string_parser(s_entry * type, char **value,
                         service_h * from_service)
{
    char *to;

    assert(from_service);
    assert(*value);
    assert(value);

    D_("string_parser(%s,s,%s);\n", type->opt_name, from_service->name);
    if ((to = st_dup_line(value)))
    {
        service_db_set_string(type, from_service, to);
        return (TRUE);
    }
    F_("Unable to parse line!\n");
    return (FALSE);
}

static int int_parser(s_entry * type, char **value, service_h * from_service)
{
    char *to;

    assert(from_service);
    assert(*value);
    assert(value);

    D_("int_parser(%s,s,%s);\n", type->opt_name, from_service->name);
    if ((to = st_dup_line(value)))
    {
        service_db_set_int(type, from_service, atoi(to));
        if (to)
            free(to);
    }
    return (TRUE);
}

static int strings_parser(s_entry * type, char **value,
                          service_h * from_service)
{
    char *to = NULL;

    assert(from_service);
    assert(*value);
    assert(value);

    D_("strings_parser(%s,s,%s);\n", type->opt_name, from_service->name);

    while ((to = st_dup_next_word(value)))
    {
        service_db_set_another_string(type, from_service, to);
        JUMP_SPACES(*value);
    }
    return (TRUE);
}

static int script_parser(s_entry * type, char **value,
                         service_h * from_service)
{
    int stack_length = 0;

    assert(from_service);
    assert(*value);
    assert(value);

    D_("script_parser(%s,s,%s);\n", type->opt_name, from_service->name);
    stack_length = chars_to_end_of_this_stack(*value);

    service_db_set_string(type, from_service,
                          i_strndup(*value, stack_length));
    (*value) = (*value) + stack_length;
    return (TRUE);
}






int module_init(const char *version)
{

    D_("i_parser: module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    return (initng_add_hook(PARSERS, 50, &initng_i_parser));

}

void module_unload(void)
{
    D_("i_parser: module_unload();\n");
    initng_del_hook(PARSERS, &initng_i_parser);
}
