
/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"


#include <stdio.h>
#include <stdlib.h>                         /* free() exit() */
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_static_data_id.h"

#include "initng_reload.h"

#define SAVE_FILE "/tmp/active_db_save.v8"

s_command FAST_RELOAD = { 'c', "hot_reload", TRUE_OR_FALSE_COMMAND, NO_OPT,
    {(void *) &cmd_fast_reload},
    "Fast Reload"
};



static void read_adb_file(void)
{
    FILE *fil;
    active_h *new_entry = NULL;
    data_save_struct entry;
    s_data *d;
    int i;
    int pnr = 0;
    process_h *process = NULL;

    fil = fopen(SAVE_FILE, "r");

    if (!fil)
        return;

    while (!feof(fil))
    {
        if (!fread(&entry, sizeof(entry), 1, fil))
            continue;

        if (active_db_find_by_name(entry.name))
        {
            W_("Entry exists, won't create it!\n");
            continue;
        }

        if (!(new_entry = active_db_new(i_strdup(entry.name))))
        {
            F_("Can't create new active!\n");
            continue;
        }
        new_entry->a_status = entry.a_status;
        new_entry->type = entry.type;
        new_entry->time_got_status = entry.time_got_status;

	/* walk thru all processes */
	pnr=0;
	while(entry.process[pnr].ptype != T_UNKNOWN && pnr < MAX_PROCESSES)
	{
	    /* allocate the process */
            process = process_db_new(entry.process[pnr].ptype);
	    if(!process) continue;
	    
	    /* fill the data */
	    process->pid = entry.process[pnr].pid;
	    process->out_pipe[0] = entry.process[pnr].stdout1;
	    process->out_pipe[1] = entry.process[pnr].stdout2;
	    process->r_code = entry.process[pnr].rcode;
	    
	    /* allocate a new output buffer for this process */
	    process->buffer_pos=0;
	    process->buffer=NULL;
	    
	    /* add this process to the list */
	    list_add(&process->list, &new_entry->processes.list);
	    
	    printf("Added process type %i to %s\n", process->pt, new_entry->name);
	    
	    pnr++;
	}

        i = 0;
        while (entry.data[i].opt_type)
        {
            d = (s_data *) initng_calloc(1, sizeof(s_data));
            d->type = initng_sdt_find(entry.data[i].type);
            if (!d->type)
            {
                F_("Did not found %s!\n", entry.data[i].type);
                free(d);
                i++;
                continue;
            }
            switch (d->type->opt_type)
            {
                case STRING:
                case STRINGS:
                    d->t.s = i_strdup(entry.data[i].t.s);
                    break;
                case INT:
                    d->t.i = entry.data[i].t.i;
                    break;
                default:
                    break;
            }
            list_add(&d->list, &new_entry->data.list);
            i++;
        }
        if (active_db_add(new_entry) != TRUE)
        {
            F_("Could not add entry!\n");
            active_db_free(new_entry);
            continue;
        }
    }

    fclose(fil);
    if (unlink(SAVE_FILE) != 0)
    {
        W_("Failed removing file " SAVE_FILE " !!!\n");
        return;
    }
}

static void cmd_fast_reload(char *arg, FILE * fd)
{
    FILE *fil;
    active_h *current, *q = NULL;
    data_save_struct entry;
    process_h *process = NULL;
    int i;
    int pnr=0;
    s_data *c_d = NULL;

    (void) arg;

    fil = fopen(SAVE_FILE, "w+");

    /* walk the active_db */
    while_active_db_safe(current, q)
    {
        if (strlen(current->name) >= MAX_SERVICE_NAME_STRING_LEN)
        {
            F_("Service name is to long, it won't fit the fileformat spec! max is %i, won't save this service!\n", MAX_SERVICE_NAME_STRING_LEN);
            continue;
        }
        strcpy(entry.name, current->name);
        entry.a_status = current->a_status;
        entry.type = current->type;
        entry.time_got_status = current->time_got_status;

        /* collet some proceses */
        process = NULL;
	pnr=0;
        while_processes(process, current)
        {
	    entry.process[pnr].ptype = process->pt;
	    entry.process[pnr].pid = process->pid;
	    entry.process[pnr].stdout1 = process->out_pipe[0];
	    entry.process[pnr].stdout2 = process->out_pipe[1];
	    entry.process[pnr].rcode = process->r_code;
	    pnr++;
	    if(pnr >= MAX_PROCESSES)
		break;
        }
	entry.process[pnr].ptype = T_UNKNOWN;

        /* reset data */
        for (i = 0; i < MAX_ENTRYS_FOR_SERVICE; i++)
        {
            entry.data[i].opt_type = 0;
            entry.data[i].type[0] = '\0';
            entry.data[i].t.i = 0;
        }

        i = 0;
        list_for_each_entry(c_d, &current->data.list, list)
        {
            if (!c_d->type)
                continue;
            if (!c_d->type->opt_name)
                continue;

            entry.data[i].opt_type = c_d->type->opt_type;
            strncpy(entry.data[i].type, c_d->type->opt_name,
                    MAX_TYPE_STRING_LEN);

            /* copy the data */
            switch (c_d->type->opt_type)
            {
                case STRING:
                case STRINGS:
                    strncpy(entry.data[i].t.s, c_d->t.s, MAX_DATA_STRING_LEN);
                    break;
                case INT:
                    entry.data[i].t.i = c_d->t.i;
                    break;
                default:
                    break;
            }

            /* save in next data entry row */
            i++;

            /* maximum data entrys that can be saved */
            if (i == MAX_ENTRYS_FOR_SERVICE)
            {
                F_("Maximum 20 data entries / service can't be saved!\n");
                break;
            }
        }

        printf("Saving : %s\n", entry.name);
        fwrite(&entry, sizeof(entry), 1, fil);
    }

    fclose(fil);

    if (!g.i_am_init)
        _exit(0);

    {

        char *new_argv[] =
            { i_strdup("/sbin/initng"), i_strdup("--hot_reload"), NULL };
        execve("/sbin/initng", new_argv, environ);
    }
    F_("Failed to reload initng!\n");
}

int module_init(const char *version)
{
    D_("module_init(reload);\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    read_adb_file();
    initng_command_add(&FAST_RELOAD);
    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_command_del(&FAST_RELOAD);
}
