
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <unistd.h>                         /* execv() pipe() usleep() pause() chown() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* free() exit() */
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <pwd.h>
#include <sys/stat.h>
#include <assert.h>

#include "../../src/initng_active_db.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_execute.h"
#include "../../src/initng_string_tools.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_common.h"
#include "../../src/initng_signal.h"
#include "../../src/initng_fork.h"

#include "initng_simple_launcher.h"

s_entry DAEMON = { "daemon", STRINGS, 6, DAEMON_OPT };
s_entry DAEMON_ARGS = { "daemon_args", STRING, 11, DAEMON_OPT };
s_entry KILL = { "kill", STRINGS, 4, DAEMON_OPT };
s_entry KILL_ARGS = { "kill_args", STRING, 9, DAEMON_OPT };
s_entry START = { "start", STRINGS, 5, SERVICE_OPT };
s_entry START_ARGS = { "start_args", STRING, 10, SERVICE_OPT };
s_entry STOP = { "stop", STRINGS, 4, SERVICE_OPT };
s_entry STOP_ARGS = { "stop_args", STRING, 9, SERVICE_OPT };


static int simple_exec(process_h * p, active_h * s, char *e, char *ea);

static int initng_s_launch(active_h * service, process_h * process)
{
    char *e = NULL;
    char *t = NULL;

    s_entry *EXEC = NULL;
    s_entry *EXEC_ARGS = NULL;

    assert(service);
    assert(service->name);

    /* set variables depending on launch type */
    switch (process->pt)
    {
        case T_START:
            EXEC = &START;
            EXEC_ARGS = &START_ARGS;
            break;
        case T_STOP:
            EXEC = &STOP;
            EXEC_ARGS = &STOP_ARGS;
            break;
        case T_DAEMON:
            EXEC = &DAEMON;
            EXEC_ARGS = &DAEMON_ARGS;
            break;
        case T_KILL:
            EXEC = &KILL;
            EXEC_ARGS = &KILL_ARGS;
            break;
        default:
            F_("Don't know how to handle launch type %i.\n", process->pt);
            return (FALSE);
    }



    /* WE ARE EXECUTING A START FILE */
    if (active_db_is(EXEC, service))
    {
        while ((e = active_db_get_next_string(EXEC, service, e)))
        {

            struct stat test;

            t = fix_variables(e, service);
            if (stat(t, &test) == -1)
            {
                F_("Stat failed on %s\n", t);
                free(t);
                continue;

            }

            if (simple_exec(process, service, t,
                            active_db_get_string(EXEC_ARGS, service)))
            {
                free(t);
                return (TRUE);
            }
            D_("%s did not work\n", t);
            free(t);
        }
        return (FAIL);
    }

    return (FALSE);
}


static int simple_exec(process_h * process_to_exec, active_h * s,
                       char *e, char *ea)
{

    /* called from inside the service directory, return the PID or 0 on error */

    /* This is the real service kicker */
    pid_t pid_fork;             /* pid got from fork() */
    char **argv = 0;            /* Arguments */
    int argc;                   /* number of arguments */

    assert(process_to_exec);
    assert(s);
    assert(s->name);
    assert(e);

    if ((pid_fork = initng_fork(s, process_to_exec)) == 0)
    {
        char *exec = NULL;
        char *exec_args = NULL;

        ea = fix_variables(ea, s);

        exec = e;
        exec_args = ea;

        D_("FROM_FORK simple_exec(%s,%s);\n", e, ea);

        /* get args from service database */
        if (exec_args)
        {
            D_("FROM_FORK (%s): start args: %s\n", exec, exec_args);

            /*
             * obs!, the split_n, creates a aray, of pointers that points
             * to exec_args, if exec_args is free, argv is usless!
             */
            argv = split_n(exec_args, &argc, 2, 1);

            /* i don't know about this */
            if (!argv[argc - 1])
            {
                argv[argc - 1] = '\0';      /* the last */
            }
            else
            {
                argv[argc] = '\0';          /* must be null */
            }
        }
        else
        {
            argv = initng_calloc(2, sizeof(char *));    /* alloc a null */
            argv[1] = '\0';                 /* set to null */
        }

        /* set argv[0] */
        argv[0] = exec;


        {
            int i;

            for (i = 0; argv[i]; i++)
                D_("i.e.f.a.e.(): argv[%i]=\"%s\"\n", i, argv[i]);
        }


        /* FINALLY EXECUTE *//* execve replaces the running process */
        execve(argv[0], argv, new_environ(s));

        /* Will never get here if execve succeeded */
        F_("EEEEEEEEEEEEEEEEEEERRRRRRRRRRRRRRRRRRRRRRRRRRROOOOOOOOOOOOOOOOOOOOOOOOOORRRRRRRRRRRRRRR!!!!!!!!!\n");
        F_("Cant execute source %s, \"%s\" !\n", e, ea);

        /* free variable, good for making valgrind quit */
        /*free(argv[0]);
           free(argv);
           free(exec_args); */

        /* clean up */
        /*initng_free(); */

        /* exit 1, to mark a failure */
        _exit(1);
    }

    /* save pid of fork */
    D_("FROM_FORK Forkstarted pid %i.\n", pid_fork);

    if (pid_fork > 0)
    {
        process_to_exec->pid = pid_fork;
        return (TRUE);
    }

    process_to_exec->pid = 0;
    return (FALSE);
    /* if to test want to lock this up util fork is done ...   waitpid(pid_fork,0,0); */
}                                           /* end fork_and_exec() */


int module_init(const char *version)
{
    D_("initng_simple_plugin: module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_add_hook(LAUNCH, 40, &initng_s_launch);

    initng_sdt_add(&DAEMON);
    initng_sdt_add(&DAEMON_ARGS);
    initng_sdt_add(&START);
    initng_sdt_add(&START_ARGS);
    initng_sdt_add(&STOP);
    initng_sdt_add(&STOP_ARGS);
    initng_sdt_add(&KILL);
    initng_sdt_add(&KILL_ARGS);

    return (TRUE);
}

void module_unload(void)
{
    D_("initng_simple_plugin: module_unload();\n");

    initng_sdt_del(&DAEMON);
    initng_sdt_del(&DAEMON_ARGS);
    initng_sdt_del(&START);
    initng_sdt_del(&START_ARGS);
    initng_sdt_del(&STOP);
    initng_sdt_del(&STOP_ARGS);
    initng_sdt_del(&KILL);
    initng_sdt_del(&KILL_ARGS);

    initng_del_hook(LAUNCH, &initng_s_launch);

}
