/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"


#include <stdio.h>
#include <stdlib.h>                         /* free() exit() */
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"

#include "initng_stdout.h"

s_entry STDOUT = { "stdout", STRING, 6, STD_OPT };
s_entry STDERR = { "stderr", STRING, 6, STD_OPT };
s_entry STDALL = { "stdall", STRING, 6, STD_OPT };



static int setup_output(active_h * s, process_h * p __attribute__ ((unused)))
{

    /* string containing the filename of output */
    char *s_stdout = NULL;
    char *s_stderr = NULL;
    char *s_stdall = NULL;

    /* filedescriptors */
    int fd_stdout = -1;
    int fd_stderr = -1;
    int fd_stdall = -1;

    /* assert some */
    assert(s);
    assert(s->name);
    assert(p);

    S_;

    /* get strings if present */
    s_stdout = active_db_get_string(&STDOUT, s);
    s_stderr = active_db_get_string(&STDERR, s);
    s_stdall = active_db_get_string(&STDALL, s);

    if (!s_stdout && !s_stderr && !s_stdall)
    {
        D_("This plugin wont do anything, becouse no opt set!\n");
        return (TRUE);
    }

    /* if stdout points to same as stderr, set s_stdall */
    if (s_stdout && s_stderr && strcmp(s_stdout, s_stderr) == 0)
    {
        s_stdall = s_stdout;
        s_stdout = NULL;
        s_stderr = NULL;
    }

    /* if stdall string is set */
    if (s_stdall)
    {
        /* output all to this */
        fd_stdall =
            open(s_stdall, O_WRONLY | O_NOCTTY | O_CREAT | O_APPEND, 0644);
    }
    else
    {
        /* else set them to diffrent files */
        if (s_stdout)
            fd_stdout =
                open(s_stdout, O_WRONLY | O_NOCTTY | O_CREAT | O_APPEND,
                     0644);
        if (s_stderr)
            fd_stderr =
                open(s_stderr, O_WRONLY | O_NOCTTY | O_CREAT | O_APPEND,
                     0644);
    }

    /* print fail messages, if the files did not open */
    if (s_stdall && fd_stdall < 2)
        F_("StdALL: %s, fd %i\n", s_stdall, fd_stdall);
    if (s_stdout && fd_stdout < 2)
        F_("StdOUT: %s, fd %i\n", s_stdout, fd_stdout);
    if (s_stderr && fd_stderr < 2)
        F_("StdERR: %s, fd %i\n", s_stderr, fd_stderr);

#ifdef DEBUG
    if (s_stdall)
        D_("StdALL: %s, fd %i\n", s_stdall, fd_stdall);
    if (s_stdout)
        D_("StdOUT: %s, fd %i\n", s_stdout, fd_stdout);
    if (s_stderr)
        D_("StdERR: %s, fd %i\n", s_stderr, fd_stderr);
#endif

    /* if fd succeded to open */
    if (fd_stdall > 2)
    {
        /* dup both to stdall */
        dup2(fd_stdall, 1);
        dup2(fd_stdall, 2);
    }
    else
    {
        /* else dup them to diffrent fds */
        if (fd_stdout > 2)
            dup2(fd_stdout, 1);
        if (fd_stderr > 2)
            dup2(fd_stderr, 2);
    }

    return (TRUE);
}



int module_init(const char *version)
{
    S_;

    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_sdt_add(&STDOUT);
    initng_sdt_add(&STDERR);
    initng_sdt_add(&STDALL);

    initng_add_hook(A_FORK, 30, &setup_output);
    return (TRUE);
}

void module_unload(void)
{
    S_;

    D_("module_unload();\n");

    initng_sdt_del(&STDOUT);
    initng_sdt_del(&STDERR);
    initng_sdt_del(&STDALL);

    initng_del_hook(A_FORK, &setup_output);
}
