
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <unistd.h>                         /* execv() pipe() usleep() pause() chown() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* free() exit() */
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <pwd.h>
#include <sys/stat.h>

#include "initng_global.h"
#include "initng_process_db.h"
#include "main.h"
#include "initng_string_tools.h"            /* split() */
#include "initng_execute.h"
#include "initng_active_db.h"
#include "initng_plugin_callers.h"
#include "initng_toolbox.h"
#include "initng_common.h"
#include "initng_static_data_id.h"
#include "initng_plugin.h"

int launch(active_h * service, e_pt type)
{
    s_call *current, *q = NULL;
    process_h *process = NULL;
    int ret = 0;

    assert(service);
    D_("start_launch(%s);\n", service->name);

    /* check so there don't exists any duplents of this one */
    while_processes(process, service)
    {
        if (process->pt == type)
        {
            F_("service %s, there exists a %i process allredy!\n",
               service->name, type);
            return (FAIL);
        }
    }

    process = process_db_new(type);
    if (!process)
    {
        F_("start_launch(%s): unable to allocate start_process!\n",
           service->name);
        return (FALSE);
    }

    /* add the process to our service */
    list_add(&process->list, &service->processes.list);

    /* walk the db with hooks */
    while_call_db_safe(current, q)
    {
        if (current->ptype != LAUNCH)
            continue;

        /* call the start */
        ret = (*current->c.launch) (service, process);

        /* if lauch succeded */
        if (ret >= TRUE)
            return (TRUE);

        /* if that launcher did not get any to launch, take next launcher */
        if (ret == FALSE)
            continue;
        if (ret < FALSE)
            break;
    }

    if (ret < FALSE)
        F_("initng_execute(%s): FAILED LAUNCHING, returned FAIL\n",
           service->name);
    else
        D_("initng_execute(%s): FAILED LAUNCHING, resturned FALSE\n",
           service->name);

    /* on failure remove the process from list, and free it */
    list_del(&process->list);
    process_db_free(process);

    return (ret);
}



/* creates a custom environment variable */
char **new_environ(active_h * s)
{
    int allocate;
    int nr = 0;
    char **env;

    /* Better safe than sorry... */
    if (s && s->name == NULL)
        s->name = i_strdup("unknown");


    /*
     * FIRST, try to figure out how big array we want to create.
     */

    /*  At least 11 allocatons below, and place for plugin added to (about 100) */
    allocate = 112;

    /* count existing env's */
#ifdef USE_OLD_ENV
    while (environ[nr])
        allocate++;
#endif

    /* count ENVs in service ENV variable */
    if (s)
        allocate += active_db_count_type(&ENV, s);


    /* finally allocate */
    env = (char **) initng_calloc(allocate, sizeof(char *));

    /* duplicate */
#ifdef USE_OLD_ENV
    for (nr = 0; environ[nr] && nr < allocate; nr++)
    {
        env[nr] = i_strdup(environ[nr]);
    }
#endif

    /* add custom */
    if (nr + 5 < allocate)
    {
        env[nr++] = i_strdup("INITNG=" INITNG_VERSION);
        env[nr++] = i_strdup
            ("PATH=/bin:/usr/bin:/sbin:/usr/sbin:/usr/local/bin:/opt/bin");
        env[nr++] = i_strdup("HOME=/root");
        env[nr++] = i_strdup("USER=root");
        env[nr++] = i_strdup("TERM=linux");
    }

    if (s && (nr + 4) < allocate)
    {
        env[nr] =
            (char *) initng_calloc(1, sizeof(char) * (9 + strlen(s->name)));
        strcpy(env[nr], "SERVICE=");
        strcat(env[nr], s->name);
        nr++;

        env[nr] =
            (char *) initng_calloc(1, sizeof(char) * (6 + strlen(s->name)));
        strcpy(env[nr], "NAME=");
        strcat(env[nr], st_strip_path(s->name));
        nr++;

        env[nr] =
            (char *) initng_calloc(1, sizeof(char) * (7 + strlen(s->name)));
        strcpy(env[nr], "CLASS=");
        strcat(env[nr], s->name);
        st_strip_end(&env[nr]);
        nr++;


        env[nr] = NULL;
        /* insert all env strings from config */
        if ((env[nr] = active_db_get_next_string(&ENV, s, NULL)))
        {
            nr++;
            while (((nr + 1) < allocate) && (env[nr] =
                                             active_db_get_next_string(&ENV,
                                                                       s,
                                                                       env[nr
                                                                           -
                                                                           1])))
                nr++;

        }
    }

    /* null last */
    if (env[nr] != NULL)
        env[nr++] = NULL;


    if(s) {
        s_call *current, *safe = NULL;

        while_call_db_safe(current, safe)
        {
            if (current->ptype != NEW_ENVIRON)
                continue;
            if (nr >= allocate)
                break;

            env = (*current->c.new_environ) (s, env, (allocate - nr));

            /* if the plugin broke things */
            if (!env)
                return (NULL);

            /* count how many extra we got. */
            while (env[nr])
                nr++;
        }
    }


#ifdef DEBUG
    for (nr = 0; env[nr]; nr++)
    {
        D_("environ[%i] = \"%s\"\n", nr, env[nr]);
    }
#endif

    /* return new environ */
    return (env);
}

/* this frees an environment variable */
void free_environ(char **tf)
{
    int i = 0;

    D_("free_environ();\n");
    assert(tf);
    for (i = 0; tf[i]; tf++)
    {
        free(tf[i]);
    }
    free(tf);
}
