
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "initng.h"
#include "initng_global.h"
#include "initng_load_module.h"
#include "initng_toolbox.h"


/* HERE IS THE GLOBAL DEFINED STRUCT, WE OFTHEN RELATE TO */
s_global g;

void initng_new(int argc, char *argv[], char *env[])
{
    int i;
    
    assert(argv);
    assert(env);

    /* we want to keep a copy of the arguments passed to us */
    g.Argc = argc;
    g.Argv0 = argv[0];
    g.Argv = (char **) initng_calloc(argc + 1, sizeof(char *));
    assert(g.Argv);
    
    for (i = 0; i < argc; i++)
    {
        g.Argv[i] = i_strdup(argv[i]);
	assert(g.Argv[i]);
    }
    g.Argv[argc] = NULL;

    /* 
     * We want to change our process name, because it looks nice in 
     * "ps" output. But we can only use space that the kernel 
     * allocated for us when we started. This is argv[] and env[] 
     * together. 
     */
    g.maxproclen = 0;
    for (i = 0; i < argc; i++)
    {
        g.maxproclen += strlen(argv[i]) + 1;
    }
    for (i = 0; env[i] != NULL; i++)
    {
        g.maxproclen += strlen(env[i]) + 1;
    }
    D_("Maximum length for our process name is %d\n", g.maxproclen);

    /* initziate all databases */
    INIT_LIST_HEAD(&g.active_database.list);
    INIT_LIST_HEAD(&g.service_cache.list);
    INIT_LIST_HEAD(&g.call_db.list);
    INIT_LIST_HEAD(&g.module_db.list);
    INIT_LIST_HEAD(&g.option_db.list);
    INIT_LIST_HEAD(&g.command_db.list);

    /* default global variables */
    g.interrupt = FALSE;
    g.sys_state = STATE_NULL;
    g.runlevel = NULL;
    g.modules_to_unload = FALSE;

    /* Add defaults (static) to data_types */
    initng_sdt_add_defaults();


#ifdef DEBUG
    g.verbose = 0;
    {
        int i;

        for (i = 0; i < MAX_VERBOSES; i++)
            g.verbose_this[i] = NULL;
    }
#endif
}

void initng_free(void)
{
    int i;

    /* free all databases */
    /*history_free_all(); *//* flush history log db */
    active_db_free_all();                   /* clean process_db */
    service_db_free_all();                  /* clean service_cache_db */
    initng_sdt_free_all();                  /* clean option_db */
    initng_command_free_all();              /* clean command_db */

    /* free runlevelname string */
    if (g.runlevel)
        free(g.runlevel);

    /* free ur copy of argv, and argc */
    for (i = 0; i < g.Argc; i++)
    {
        free(g.Argv[i]);
    }
    free(g.Argv);


    /* close all open fd, over 3, this may hang boot */
    for (i = 3; i <= 1013; i++)
    {
        close(i);
    }
}
