/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef INITNG_PLUGIN
#define INITNG_PLUGIN
#include <stdarg.h>
#include "initng_system_states.h"
#include "initng_list.h"



typedef struct
{
    void (*call_module) (void);
    int fds;
} f_module_h;

typedef struct
{
    void (*call_module) (void);
    time_t seconds;
    time_t next_hit;
} t_module_h;


typedef enum
{
    EMPTY_UNKNOWN_TYPE = 0,
    ASTATUS_CHANGE = 1,
    CONFIRM_ASTATUS_CHANGE = 2,
    PARSERS = 3,
    SWATCHERS = 4,
    FDWATCHERS = 5,                         /* File descriptor watcher */
    PIPEWATCHERS = 6,                       /* stdout pipe watcher (for every process) */
    TIMERS = 7,
    MAIN = 10,
    A_FORK = 11,
    HANDLE_KILLED = 12,
    COMPENSATE_TIME = 13,
    ERR_MSG = 14,
    NEW_ENVIRON = 15,
    LAUNCH = 16,
    DEP_ON = 17
} e_ptype;


typedef union
{
    int (*test_status) (active_h * service, e_a_status status);
    int (*status_change) (active_h * service);
    service_h *(*parser) (char *name);
    void (*swatcher) (h_sys_state state);
    void (*pipewatcher) (active_h * service, int datalen, process_h * process,
                         int forceflush);
    int (*launch) (active_h * service, process_h * process);
    int (*af_launcher) (active_h * service, process_h * process);
    int (*handle_killed) (active_h * service, process_h * process);
    void (*main) (void);
    void (*compensate_time) (int seconds);
    void (*err) (e_mt mt, const char *file, const char *func, int line,
                 const char *format, va_list ap);
    char **(*new_environ) (active_h * service, char **environ,
                           int no_of_allocs_left);
    int (*dep_on_check) (active_h * service, active_h * check);
    f_module_h *fdh;
    t_module_h *tdh;

    void *pointer;
} uc;


typedef struct s_callers_s s_call;
struct s_callers_s
{
    char *from_file;
    e_ptype ptype;
    uc c;
    int order;
    struct list_head list;
};

#define while_call_db(current) list_for_each_entry_prev(current, &g.call_db.list, list)
#define while_call_db_safe(current, safe) list_for_each_entry_prev_safe(current, safe, &g.call_db.list, list)

#endif
